#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <tchar.h>
#include <direct.h>

/*
To compile this source code you'll need "blast.c" and "blast.h" files from zlib package.
Get this package from official site: http://zlib.net/
Requered files can be found in /contrib/blast/ folder inside archive.
*/

/* VS VC++ 6.0 size optimizing */
#pragma comment(linker,"/opt:nowin98")
#pragma comment(linker,"/subsystem:console")

#include "blast.h"

/* next two define and routines from "blast.c" file */

#define CHUNK 16384
#define local static /* for local function definitions */

local unsigned inf(void *how, unsigned char **buf)
{
    static unsigned char hold[CHUNK];

    *buf = hold;
    return fread(hold, 1, CHUNK, (FILE *)how);
}

local int outf(void *how, unsigned char *buf, unsigned len)
{
    return fwrite(buf, 1, len, (FILE *)how) != len;
}

#pragma pack (1)
typedef struct {
    unsigned int offs; /* absolute offset to file in archive */
    unsigned int pack; /* 0 - unpacked; 1 - packed with PKWARE implode */
    unsigned int unsz; /* unpacked size; 0 if not packed */
    unsigned int pksz; /* packed size (size in archive); same as filesize if not packed */
} tpack;

#define GET_UINT(x) ((unsigned int *)x)[0]

char KEY_DEMO[] = "NSIARKPRQPHBTE50GRIH3AYXJP2AMF3FCEYAVQO5QGA0JGIIH2AYXKVOA1VOGGU5GSQKKYEOIAQG1XRX0J4F5OEAEFI4DD3LL45VJTVOA1VOGGUKE50GRI";
char KEY_FULL[] = "AVQF3FCKE50GRIAYXJP2AMEYO5QGA0JGIIH2NHBTVOA1VOGGU5H3GSSIARKPRQPQKKYEOIAQG1XRX0J4F5OEAEFI4DD3LL45VJTVOA1VOGGUKE50GRIAYX";

unsigned char xordetect(unsigned int data){
  /* demo - plain */
  if(data == 0){
    return(0);
  }
  /* demo - key 1 */
  if(data == GET_UINT(KEY_DEMO)){
    return(1);
  }
  /* full - key 2 */
  if(data == GET_UINT(KEY_FULL)){
    return(2);
  }
 /* not a .PAK archive / unknown key */
  return(255);
}

int main(int argc, char *argv[]){
FILE *fl, *fout;
unsigned int i, sz;
unsigned char *buff;
char *toc, *htoc, *etoc, *dir, *fname;
tpack *pack;
unsigned int len;

  printf("Arx Fatalis .PAK unpacker v1.2\n(c) CTPAX-X Team 2010-2012\nhttp://www.CTPAX-X.ru/\n\n");
  if(argc != 2){
    printf("Usage: afunpak filename.pak\n\n");
    return(1);
  }
  fl = fopen(argv[1], "rb");
  if(fl == NULL){
    printf("Can\'t open \"%s\".\n", argv[1]);
    return(2);
  }
  /* read offset to TOC */
  fread(&sz, 4, 1, fl);
  fseek(fl, 0, SEEK_END);
  if(ftell(fl) - 4 <= sz){
    fclose(fl);
    printf("Not a .PAK archive.\n");
    return(3);
  }
  fseek(fl, sz, SEEK_SET);
  /* TOC size */
  fread(&sz, 4, 1, fl);
  toc = (char *) malloc(sz);
  htoc = toc;
  /* read TOC header */
  if(fread(toc, 1, sz, fl) != sz){
    free(toc);
    fclose(fl);
    printf("Not a .PAK archive.\n");
    return(3);
  }
  /* check archive file */
  i = xordetect(GET_UINT(toc));
  if(i == 255){
    free(toc);
    fclose(fl);
    printf("Not a .PAK archive or unknown xor key.\n");
    return(4);
  }
  /* check if TOC is crypted */
  if(i != 0){
    /* decrypt TOC */
    dir = ((i == 1) ? KEY_DEMO : KEY_FULL);
    len = strlen(dir);
    for(i = 0; i < sz; i++){
      toc[i] = toc[i] ^ dir[i % len];
    }
    /* --- here you can dump toc[] to file for better understanding TOC structure --- */
  }
  /* calculate end of TOC pointer */
  etoc = toc;
  etoc += sz;
  while(toc < etoc){
    /* create folder */
    dir = toc;
    len = strlen(dir) + 1;
    mkdir(dir);
    /* skip ASCIIZ folder name */
    toc += strlen(toc) + 1;
    /* number of files inside this folder */
    memmove(&sz, toc, 4);
    toc += 4;
    while(sz > 0){
      i = strlen(toc);
      fname = (char *) malloc(len + i);
      sprintf(fname, "%s%s", dir, toc);
      toc += i + 1;
      pack = (tpack *) toc;
      toc += sizeof(tpack);
      printf("%s", fname);
      fout = fopen(fname, "wb");
      free(fname);

      fseek(fl, pack->offs, SEEK_SET);
      if(pack->pack == 0){
        /* unpacked */
        buff = (unsigned char *) malloc(pack->pksz);
        fread(buff, 1, pack->pksz, fl);
        fwrite(buff, 1, pack->pksz, fout);
        free(buff);
      }else{
        /* implode */
        blast(inf, fl, outf, fout);
      }

      fclose(fout);
      printf("\n");
      sz--;
    }
  }
  free(htoc);
  fclose(fl);
  return(0);
}
